package aprioriUtil;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

public class AssociationRuleGenerator<I> {

    public List<AssocRule<I>> mineAssociationRules(FreqItemSet<I> data,
            double minConf) {
        Objects.requireNonNull(data, "The frequent itemset data is null.");

        Set<AssocRule<I>> resultSet = new HashSet<>();

        for (Set<I> itemset : data.getFreqItemList()) {
            if (itemset.size() < 2) {
                continue;
            }
            Set<AssocRule<I>> basicAssociationRuleSet = generateBasicAssocRules(itemset, data);
            generateAssocRules(itemset,basicAssociationRuleSet,data,minConf,resultSet);
        }
        return resultSet.stream().sorted((a1, a2) -> Double.compare(a2.getConf(), a1.getConf()))
                .collect(Collectors.toList());
    }

    private void generateAssocRules(Set<I> itemset,
            Set<AssocRule<I>> ruleSet,
            FreqItemSet<I> data,
            double minConf,
            Set<AssocRule<I>> collector) {
        if (ruleSet.isEmpty()) {
            return;
        }
        int size = itemset.size();
        int ruleSetSize = ruleSet.iterator().next().getCons().size();
        if (size > ruleSetSize + 1) {
            Set<AssocRule<I>> nextRules = moveOneItemToCons(itemset, ruleSet, data);
            Iterator<AssocRule<I>> ruleIterator = nextRules.iterator();
            while (ruleIterator.hasNext()) {
                AssocRule<I> rule = ruleIterator.next();
                if (rule.getConf() >= minConf) {
                    collector.add(rule);
                } else {
                    ruleIterator.remove();
                }
            }
            generateAssocRules(itemset,
                    nextRules,
                    data,
                    minConf,
                    collector);
        }
    }

    private Set<AssocRule<I>> moveOneItemToCons(Set<I> itemset,
            Set<AssocRule<I>> ruleSet,
            FreqItemSet<I> data) {
        Set<AssocRule<I>> output = new HashSet<>();
        Set<I> ante = new HashSet<>();
        Set<I> cons = new HashSet<>();
        double itemSupCount = data.getSupportCountMap().get(itemset);
        for (AssocRule<I> rule : ruleSet) {
            ante.clear();
            cons.clear();
            ante.addAll(rule.getAnte());
            cons.addAll(rule.getCons());
            for (I item : rule.getAnte()) {
                ante.remove(item);
                cons.add(item);
                int antecedentSupportCount = data.getSupportCountMap()
                        .get(ante);
                AssocRule<I> newRule = new AssocRule<>(
                        ante,
                        cons,
                        itemSupCount / antecedentSupportCount);

                output.add(newRule);
                ante.add(item);
                cons.remove(item);
            }
        }

        return output;
    }

    private Set<AssocRule<I>> generateBasicAssocRules(Set<I> itemset,
            FreqItemSet<I> data) {
        Set<AssocRule<I>> basicAssocRuleSet = new HashSet<>(itemset.size());

        Set<I> ante = new HashSet<>(itemset);
        Set<I> cons = new HashSet<>(1);

        for (I item : itemset) {
            ante.remove(item);
            cons.add(item);

            int itemsetSuppCount = data.getSupportCountMap().get(itemset);
            int anteSupCount = data.getSupportCountMap()
                    .get(ante);

            double conf = 1.0 * itemsetSuppCount
                    / anteSupCount;

            basicAssocRuleSet.add(new AssocRule(ante,cons,conf));
            ante.add(item);
            cons.remove(item);
        }

        return basicAssocRuleSet;
    }

    private void checkMinimumConfidence(double minimumConfidence) {
        if (Double.isNaN(minimumConfidence)) {
            throw new IllegalArgumentException(
                    "The input minimum confidence is NaN.");
        }

        if (minimumConfidence < 0.0) {
            throw new IllegalArgumentException(
                    "The input minimum confidence is negative: " +
                            minimumConfidence + ". " +
                            "Must be at least zero.");
        }

        if (minimumConfidence > 1.0) {
            throw new IllegalArgumentException(
                    "The input minimum confidence is too large: " +
                            minimumConfidence + ". " +
                            "Must be at most 1.");
        }
    }
}
